import * as d3 from "d3";
import { defaultTreeOptions } from "../options/tree-default-options";
import binaryTreeUtiles from "../utiles/binary-tree-utiles";
let treeOptions;
let svg;
let el;
let tree;
let selectedNode = new Map();
let isAnimationEnded = true;
let onNodeClickCallback;
let isFreeze = false;
// FIXME: use single principle to refactor below code
export default function binaryTreeDrawer() {
    function draw(element, binaryTree, options) {
        el = element;
        tree = binaryTree;
        treeOptions = getOptions(options);
        initSvg();
        let { nodes, links } = getTreeData();
        addOptionToBinaryTree(nodes);
        let binaryTreeEnter = getBinaryTreeEnter(nodes);
        if (treeOptions.drawNodes)
            drawNodes(binaryTreeEnter, treeOptions.nodeStyleOptions);
        if (treeOptions.drawConnections)
            drawConnections(links, treeOptions.linkStyleOptions);
        if (treeOptions.animation)
            animateConnections();
        if (treeOptions.drawNodevalue)
            addTextToNodes(binaryTreeEnter);
        if (treeOptions.nodeStyleOptions.onMouseHoverColor)
            addMouseHoverToNodes();
        if (treeOptions.zoom)
            addZoom();
    }
    function resetTree() {
        isAnimationEnded = true;
        removeTree();
        draw(el, tree, treeOptions);
        onNodeClick(onNodeClickCallback);
        if (selectedNode.size > 0) {
            const [node] = selectedNode.keys();
            const [color] = selectedNode.values();
            fillNode(node, color);
        }
    }
    function stopTransition() {
        d3.selectAll("path").transition();
        d3.selectAll("circle").transition();
    }
    function removeTree() {
        d3.select("#svg-container").remove();
    }
    function animatePath(pathId, pathAnimationStyle) {
        if (!isAnimationEnded)
            return;
        const id = !treeOptions?.linkStyleOptions?.addAnimationPaths
            ? "#path"
            : "#animationPath";
        const path = svg.select(`${id}${pathId}`);
        if (pathAnimationStyle?.animationClass) {
            return path.attr("class", pathAnimationStyle?.animationClass);
        }
        return path
            .transition(300)
            .duration(pathAnimationStyle?.duration)
            .attr("stroke", "#626ee3");
    }
    function animateNode(nodeId, nodeAnimationStyle) {
        if (!isAnimationEnded)
            return;
        const id = !treeOptions?.nodeStyleOptions?.addAnimationNodes
            ? "#node"
            : "#animationNode";
        const node = svg.select(`${id}${nodeId}`);
        if (nodeAnimationStyle?.animationClass) {
            return node.attr("class", nodeAnimationStyle?.animationClass);
        }
        return node
            .transition(300)
            .duration(nodeAnimationStyle?.duration)
            .attr("stroke", "#626ee3");
    }
    function setFreeze(freeze) {
        isFreeze = freeze;
    }
    function onNodeClick(callback) {
        if (typeof callback !== "function" || isFreeze) {
            return;
        }
        onNodeClickCallback = callback;
        svg.selectAll("#g").on("click", (event, node) => {
            if (isFreeze)
                return;
            if (isAnimationEnded) {
                if (treeOptions.nodeStyleOptions.selectedNodeColor) {
                    chooseNode(node);
                }
                const [currentNode] = selectedNode.keys();
                callback(currentNode, event);
            }
        });
    }
    function chooseNode(node) {
        const [currentNode] = selectedNode.keys();
        if (node === currentNode || node.data.value === currentNode?.value) {
            fillNode(node, treeOptions.nodeStyleOptions.fillCollor);
            selectedNode = new Map();
            return;
        }
        if (currentNode) {
            fillNode(currentNode, treeOptions.nodeStyleOptions.fillCollor);
            selectedNode = new Map();
        }
        fillNode(node, treeOptions.nodeStyleOptions.selectedNodeColor);
    }
    function selectNode(node) {
        if (!node)
            return;
        fillNode(node, treeOptions.nodeStyleOptions.selectedNodeColor);
    }
    function fillNode(node, color) {
        selectedNode.set(node, color);
        let id = node.nodeId ? `#${node.nodeId}` : `#node${node.value}`;
        d3.select(id).style("fill", color);
    }
    function initSvg() {
        svg = d3
            .select(`${el}`)
            .append("svg")
            .attr("id", "svg-container")
            .attr("width", getTreeWidth())
            .attr("height", treeOptions.height)
            .attr("transform", "translate(" +
            treeOptions.transformTreeFromLeft +
            "," +
            treeOptions.transformTreeFromTop +
            ")")
            .append("g")
            .attr("id", "g-container")
            .attr("transform", "translate(" + 0 + "," + 30 + ")");
    }
    function addZoom() {
        const scale = d3.zoom().scaleExtent([0.5, 5]).on("zoom", zoom);
        d3.select("svg").call(scale);
        function zoom(e) {
            d3.select("svg g").attr("transform", e.transform);
        }
    }
    function getTreeData() {
        let tree = d3.tree().size([getTreeWidth(), treeOptions.height]);
        let bst = getHierarchyBT();
        return {
            nodes: getRemovedEmptyNodes(tree(bst).descendants()),
            links: bst.links(),
        };
    }
    function getHierarchyBT() {
        return d3.hierarchy(tree.root, function (node) {
            node.children = [];
            if (node.left) {
                if (!node.right) {
                    node.children.push({});
                }
                node.children.unshift(node.left);
            }
            if (node.right) {
                if (!node.left) {
                    node.children.unshift({});
                }
                node.children.push(node.right);
            }
            return node.children;
        });
    }
    function addOptionToBinaryTree(nodes) {
        nodes.forEach((node) => {
            node.y = node.depth * treeOptions.sizeBetweenNodes;
            if (treeOptions.autoFitText) {
                const newRadius = (node.data?.value?.toString().length / 2) * 11;
                if (treeOptions.nodeStyleOptions.radius < newRadius) {
                    treeOptions.nodeStyleOptions.radius = newRadius;
                }
            }
            return node;
        });
    }
    function getRemovedEmptyNodes(nodes) {
        return nodes.filter((node) => {
            return !isEmpty(node.data);
        });
    }
    function getRemovedEmptyLinks(links) {
        return links.filter((link) => {
            return link?.target?.id;
        });
    }
    function isEmpty(obj) {
        return Object.keys(obj).length === 1;
    }
    function drawNodes(binaryTreeEnter, nodeStyleOptions) {
        binaryTreeEnter
            .append("circle")
            .attr("r", nodeStyleOptions.radius)
            .attr("fill", nodeStyleOptions.fillCollor)
            .attr("id", function (node) {
            node.nodeId = `node${node.data.value || node.data.id}`;
            return node.nodeId;
        });
        if (nodeStyleOptions.addAnimationNodes) {
            binaryTreeEnter
                .append("circle")
                .attr("r", nodeStyleOptions.radius)
                .attr("fill", nodeStyleOptions.addAnimationNodes ? "transparent" : nodeStyleOptions.fillCollor)
                .attr("id", function (node) {
                node.nodeId = `animationNode${node.data.value || node.data.id}`;
                return node.nodeId;
            });
        }
        const style = {
            stroke: nodeStyleOptions.strokeColor,
            "stroke-width": nodeStyleOptions.strokeWidth,
        };
        !nodeStyleOptions.styleClass
            ? addCssAttributesToElements(svg.selectAll("circle"), style)
            : addClassToElements("circle", nodeStyleOptions.styleClass);
    }
    function addTextToNodes(binaryTreeEnter) {
        binaryTreeEnter
            .append("text")
            .attr("x", 0)
            .attr("dy", ".35em")
            .attr("text-anchor", "middle")
            .text(function (node) {
            return node.data.value;
        });
    }
    function getBinaryTreeEnter(nodes) {
        var binaryTree = addBTNodesId(nodes);
        return binaryTree
            .enter()
            .append("g")
            .attr("cursor", "pointer")
            .attr("id", "g")
            .attr("transform", function (node) {
            return "translate(" + node.x + "," + node.y + ")";
        });
    }
    function addMouseHoverToNodes() {
        svg
            .selectAll("#g")
            .on("mouseenter", (e) => handleMouseEnter(e))
            .on("mouseleave", (e) => handleMouseLeave(e));
    }
    function handleMouseEnter(event) {
        d3.select(event.target)
            .select("circle")
            .attr("fill", treeOptions.nodeStyleOptions.onMouseHoverColor);
    }
    function handleMouseLeave(event) {
        d3.select(event.target)
            .select("circle")
            .attr("fill", treeOptions.nodeStyleOptions.fillCollor);
    }
    function addBTNodesId(nodes) {
        return svg.selectAll("circle").data(nodes, function (node, i) {
            return node.value || (node.id = i);
        });
    }
    function drawConnections(links, linkStyleOptions) {
        var connections = svg
            .selectAll("path.link")
            .data(getRemovedEmptyLinks(links), function (link) {
            return link.target.id;
        });
        var paths = addPaths(connections, "path");
        // add the same paths for animation
        if (linkStyleOptions.addAnimationPaths) {
            addPaths(connections, "animationPath");
        }
        if (treeOptions.animation) {
            addNewPropToConnects(paths);
        }
        const style = {
            stroke: linkStyleOptions.strokeColor,
            "stroke-width": linkStyleOptions.strokeWidth,
        };
        !linkStyleOptions.styleClass
            ? addCssAttributesToElements(svg.selectAll("path"), style)
            : addClassToElements("path", linkStyleOptions.styleClass);
    }
    function addPaths(connections, pathId) {
        return connections
            .enter()
            .insert("path", "g")
            .attr("id", function (e) {
            e.pathId = `${pathId}${e.target.data.value || e.target.id}`;
            return e.pathId;
        })
            .attr("fill", "none")
            .attr("d", getLinkGenarator())
            .each(function (d) {
            d.totalLength = this.getTotalLength();
        });
    }
    function addNewPropToConnects(connections) {
        connections
            .attr("stroke-dasharray", (d) => d.totalLength + " " + d.totalLength)
            .attr("stroke-dashoffset", (d) => d.totalLength);
    }
    function animateConnections(paths = getPathes(), pathIndex = 0) {
        isAnimationEnded = false;
        if (paths.length <= pathIndex) {
            isAnimationEnded = true;
            return;
        }
        const path1 = paths[pathIndex];
        const path2 = paths[pathIndex + 1];
        if (path1) {
            transitionPath(path1.id).on("end");
        }
        if (path2) {
            transitionPath(path2.id).on("end", function () {
                animateConnections(paths, pathIndex + 2);
            });
        }
    }
    function getPathes() {
        return svg.selectAll("path")._groups[0];
    }
    function transitionPath(id) {
        return svg
            .select("#" + id)
            .transition()
            .duration(treeOptions.duration)
            .ease(d3.easeLinear)
            .attr("stroke-dashoffset", 0);
    }
    function getLinkGenarator() {
        return d3
            .linkVertical()
            .x(function (link) {
            return link.x;
        })
            .y(function (link) {
            return link.y;
        });
    }
    function getTreeWidth() {
        return treeOptions.autoExpandTreeSize
            ? binaryTreeUtiles().getBinaryTreeWidth(tree)
            : treeOptions.width;
    }
    function getOptions(treeOptions = {}) {
        let options = { ...defaultTreeOptions, ...treeOptions };
        options.nodeStyleOptions = {
            ...defaultTreeOptions.nodeStyleOptions,
            ...treeOptions.nodeStyleOptions,
        };
        options.linkStyleOptions = {
            ...defaultTreeOptions.linkStyleOptions,
            ...treeOptions.linkStyleOptions,
        };
        return options;
    }
    function addCssAttributesToElements(element, styleOptions) {
        Object.keys(styleOptions).forEach((attrName) => {
            if (attrName && styleOptions[attrName])
                element.attr(attrName, styleOptions[attrName]);
        });
    }
    function addClassToElements(element, className) {
        svg.selectAll(element).attr("class", className);
    }
    return {
        draw: draw,
        onNodeClick: onNodeClick,
        animatePath: animatePath,
        animateNode: animateNode,
        stopTransition: stopTransition,
        resetTree: resetTree,
        removeTree: removeTree,
        selectNode: selectNode,
        setFreeze: setFreeze
    };
}
//# sourceMappingURL=binaryTreeDrawer.js.map